"""Visualization functions."""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats.mstats import mquantiles
from typing import List

from .constant import ColName, chrom_len


def get_qq_df(indf: pd.DataFrame) -> pd.DataFrame:
    """Convert GWAS P-values to quantiles."""
    q_pos = np.concatenate(
        [np.arange(99) / len(indf[ColName.P]), np.logspace(-np.log10(len(indf[ColName.P])) + 2, 0, 100)]
    )
    q_data = mquantiles(indf[ColName.P], prob=q_pos, alphap=0, betap=1, limit=(0, 1))
    qq_df = pd.DataFrame()
    qq_df['observed'] = -np.log10(q_data[1:])
    qq_df['expected'] = -np.log10(q_pos[1:])
    return qq_df


def qqplot(qq_df: pd.DataFrame, ax, **kwargs):
    """Plot quantile-quantile plot."""
    if 'expected' not in qq_df.columns or 'observed' not in qq_df.columns:
        qq_df = get_qq_df(qq_df)
    ax.scatter(qq_df['expected'], qq_df['observed'], color='k', **kwargs)
    ax.set_xlabel('Expected')
    ax.set_ylabel('Observed')
    ax.set_title('Quantile-Quantile Plot')
    ax.plot([0, qq_df['expected'].max()], [0, qq_df['expected'].max()], color='r', linestyle='-')
    return ax


def get_manh_df(indf: pd.DataFrame, reduce_size: float = 0.3) -> pd.DataFrame:
    """Convert GWAS sumstat to dataframe for manhattan plot."""
    agg_offset = 0
    bp_offset = {}
    for chrom, offset in chrom_len.items():
        bp_offset[chrom] = agg_offset
        agg_offset += offset
    plotdf = indf[[ColName.CHR, ColName.BP, ColName.P]].copy()
    plotdf['x'] = plotdf[ColName.BP] + plotdf[ColName.CHR].map(bp_offset)
    plotdf['y'] = -np.log10(plotdf[ColName.P])
    num_x_bins = int(20000 * reduce_size)
    num_y_bins = int(plotdf['y'].max() * 10 * reduce_size)

    x_bins = np.linspace(plotdf['x'].min(), plotdf['x'].max(), num_x_bins)
    y_bins = np.linspace(plotdf['y'].min(), plotdf['y'].max(), num_y_bins)

    plotdf['x_bin'] = np.digitize(plotdf['x'], x_bins)
    plotdf['y_bin'] = np.digitize(plotdf['y'], y_bins)

    plotdf = plotdf.drop_duplicates(subset=['x_bin', 'y_bin'])
    return plotdf


def manhattan(manh_df: pd.DataFrame, ax, colors: List[str] = ['red', 'blue'], **kwargs):
    """Plot manhattan plot."""
    if 'x_bin' not in manh_df.columns or 'y_bin' not in manh_df.columns:
        plotdf = get_manh_df(manh_df)
    else:
        plotdf = manh_df.copy()
    if len(colors) < 2:
        raise ValueError('colors must have at least 2 colors')
    plotdf['color'] = plotdf[ColName.CHR].apply(lambda x: colors[x % len(colors)])
    ax.scatter(plotdf['x'], plotdf['y'], c=plotdf['color'], **kwargs)
    chrom_label = {}
    agg_offset = 0
    for k, v in chrom_len.items():
        chrom_label[str(k)] = int(v / 2) + agg_offset
        agg_offset += v
    ax.set_xticks(list(chrom_label.values()))
    ax.set_xticklabels(list(chrom_label.keys()))
    return ax


def qqman(
    indf: pd.DataFrame, figsize: tuple = (24, 6), width_ratios: List[int] = [3, 1], colors=['red', 'blue'], **kwargs
):
    """Plot QQ and Manhattan plot."""
    from matplotlib import gridspec

    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(1, 2, width_ratios=width_ratios)
    ax0 = plt.subplot(gs[0])
    ax1 = plt.subplot(gs[1])
    qqplot(indf, ax1, **kwargs)
    manhattan(indf, ax0, colors=colors, **kwargs)
    ax0.set_xlabel('Chromosome', fontsize=14)
    ax0.set_ylabel('-$\mathregular{log_{10}}$P', fontsize=14)
    ax1.set_xlabel('Observed -$\mathregular{log_{10}}$P', fontsize=14)
    ax1.set_ylabel('Expected -$\mathregular{log_{10}}$P', fontsize=14)
    ax0.spines['right'].set_visible(False)
    ax0.spines['top'].set_visible(False)
    ax1.spines['right'].set_visible(False)
    ax1.spines['top'].set_visible(False)
    fig.tight_layout()



"""Module for annotating a file with the results of a Smunger run."""

import logging
import tabix
import pandas as pd
from subprocess import check_output
from io import StringIO
from smunger.constant import ColName
from smunger.smunger import make_SNPID_unique

logger = logging.getLogger("annotate")


def annotate_rsid(
    indf: pd.DataFrame,
    database: str,
    rsid_col: str = ColName.RSID,
    chrom_col: str = ColName.CHR,
    pos_col: str = ColName.BP,
    ea_col: str = ColName.EA,
    nea_col: str = ColName.NEA,
) -> pd.DataFrame:
    """Annotate a dataframe with rsids."""
    tb = tabix.open(database)
    chunk_df = make_SNPID_unique(indf, chrom_col, pos_col, ea_col, nea_col)
    # chunk_df = chunk_df.drop_duplicates(subset=[ColName.SNPID])
    if len(chunk_df) == 0:
        return pd.DataFrame()
    chrom = chunk_df[chrom_col].iloc[0]
    start = chunk_df[pos_col].min()
    end = chunk_df[pos_col].max()
    rsid_map = pd.DataFrame(
        columns=[ColName.CHR, ColName.BP, "rsid", "ref", "alt"],
        data=tb.query(str(chrom), start - 1, end),
    )
    rsid_map = make_SNPID_unique(rsid_map, ColName.CHR, ColName.BP, "ref", "alt")
    rsid_map = rsid_map.drop_duplicates(subset=[ColName.SNPID])
    rsid_map = pd.Series(data=rsid_map["rsid"].values, index=rsid_map[ColName.SNPID].values)  # type: ignore
    chunk_df[rsid_col] = chunk_df[ColName.SNPID].map(rsid_map)
    del chunk_df[ColName.SNPID]
    return chunk_df


def annotate_rsid_file(
    infile: str,
    outfile: str,
    database: str,
    chunksize: int = 2000000,
    rsid_col: str = ColName.RSID,
    chrom_col: str = ColName.CHR,
    pos_col: str = ColName.BP,
    ea_col: str = ColName.EA,
    nea_col: str = ColName.NEA,
) -> None:
    """Annotate a file with rsids."""
    ith = 0
    for df in pd.read_csv(infile, sep="\t", chunksize=100000):
        for chrom, chr_df in df.groupby(chrom_col):
            chr_df = chr_df.sort_values(pos_col)
            for i in range(chr_df[pos_col].min(), chr_df[pos_col].max() + 1, chunksize):
                chunk_df = chr_df[(chr_df[pos_col] >= i) & (chr_df[pos_col] < i + chunksize)].copy()
                chunk_df = annotate_rsid(
                    chunk_df, database, rsid_col, chrom_col, pos_col, ea_col, nea_col
                )
                logger.info(
                    f"Processing {chrom}:{i}-{i + chunksize}, chunk No.{ith}, {len(chunk_df)} rows."
                )
                ith += 1
                if ith == 1:
                    chunk_df.to_csv(outfile, sep="\t", index=False, mode="w")
                else:
                    chunk_df.to_csv(outfile, sep="\t", index=False, mode="a", header=False)


def update_rsid(
    indf: pd.DataFrame,
    database: str,
    rsid_col: str = ColName.RSID,
    chrom_col: str = ColName.CHR,
    pos_col: str = ColName.BP,
    ea_col: str = ColName.EA,
    nea_col: str = ColName.NEA,
) -> pd.DataFrame:
    """Update rsids in a dataframe."""
    # tb = tabix.open(database)
    pass


def annotate_alleles_from_rsid_pos(
    indf: pd.DataFrame,
    database: str,
    rsid_col: str,
    chrom_col: str,
    pos_col: str,
    a1_col: str,
    a2_col: str,
) -> pd.DataFrame:
    """Annotate a dataframe with rsids."""
    chunk_df = indf.copy()
    tb = tabix.open(database)
    chunk_df = chunk_df[
        (chunk_df[rsid_col].notnull())
        & (chunk_df[chrom_col].notnull())
        & (chunk_df[pos_col].notnull())
    ]
    if chunk_df.shape[0] == 0:
        logger.warning("No rsids found in the input file.")
        return pd.DataFrame()
    chunk_df["index"] = chunk_df.index
    chunk_df[rsid_col] = chunk_df[rsid_col].astype(str)
    chrom = chunk_df[chrom_col].iloc[0]
    start = chunk_df[pos_col].min()
    end = chunk_df[pos_col].max()
    query_res = pd.DataFrame(
        columns=["chr_dbsnp", "bp_dbsnp", "rsid_dbsnp", "ref_dbsnp", "alt_dbsnp"],
        data=tb.query(str(chrom), start - 1, end),
    )
    query_res = query_res.astype(
        {"chr_dbsnp": str, "bp_dbsnp": int, "rsid_dbsnp": str, "ref_dbsnp": str, "alt_dbsnp": str}
    )
    chunk_df = chunk_df.merge(query_res, how="left", left_on=rsid_col, right_on="rsid_dbsnp")
    n_wrong_chrom = len(
        chunk_df[
            (chunk_df[chrom_col].astype(str) != chunk_df["chr_dbsnp"])
            & (chunk_df["chr_dbsnp"].notnull())
            & (chunk_df[chrom_col].notnull())
        ]["rsid_dbsnp"].unique()
    )
    if n_wrong_chrom > 0:
        logger.warning(f"found {n_wrong_chrom} rsids with different chrom")
    n_wrong_pos = len(
        chunk_df[
            (chunk_df[pos_col] != chunk_df["bp_dbsnp"])
            & (chunk_df["bp_dbsnp"].notnull())
            & (chunk_df[pos_col].notnull())
        ]["rsid_dbsnp"].unique()
    )
    if n_wrong_pos > 0:
        logger.warning(f"found {n_wrong_pos} rsids with different pos")
    chunk_df.loc[chunk_df[a1_col].isnull() & chunk_df[a2_col].isnull(), a1_col] = chunk_df[
        "ref_dbsnp"
    ]
    chunk_df.loc[chunk_df[a1_col].isnull() & chunk_df[a2_col].isnull(), a2_col] = chunk_df[
        "alt_dbsnp"
    ]
    chunk_df.loc[
        (chunk_df[a1_col].isnull() & chunk_df[a2_col].notnull())
        & (
            (chunk_df[a2_col] == chunk_df["ref_dbsnp"])
            | (chunk_df[a2_col] == chunk_df["alt_dbsnp"])
        ),
        a1_col,
    ] = chunk_df["ref_dbsnp"].where(
        chunk_df[a2_col] == chunk_df["alt_dbsnp"], chunk_df["alt_dbsnp"]
    )
    chunk_df.loc[
        (chunk_df[a1_col].notnull() & chunk_df[a2_col].isnull())
        & (
            (chunk_df[a1_col] == chunk_df["ref_dbsnp"])
            | (chunk_df[a1_col] == chunk_df["alt_dbsnp"])
        ),
        a2_col,
    ] = chunk_df["alt_dbsnp"].where(
        chunk_df[a1_col] == chunk_df["ref_dbsnp"], chunk_df["ref_dbsnp"]
    )
    chunk_df = chunk_df.drop_duplicates(subset=["index"])
    for col in ["chr_dbsnp", "bp_dbsnp", "rsid_dbsnp", "ref_dbsnp", "alt_dbsnp", "index"]:
        del chunk_df[col]
    return chunk_df


def annotate_alleles_from_rsid_pos_file(
    infile: str,
    outfile: str,
    database: str,
    chunksize: int = 2000000,
    rsid_col: str = ColName.RSID,
    chrom_col: str = ColName.CHR,
    pos_col: str = ColName.BP,
    a1_col: str = ColName.EA,
    a2_col: str = ColName.NEA,
) -> None:
    """Annotate a file with rsids."""
    logger.warning("Annotate alleles from rsID and coordinates is not recommended!!!")
    ith = 0
    for df in pd.read_csv(infile, sep="\t", chunksize=100000):
        for chrom, chr_df in df.groupby(chrom_col):
            chr_df = chr_df.sort_values(pos_col)
            start = chr_df[pos_col].min()
            end = start + chunksize
            while not pd.isna(start):
                chunk_df = chr_df[(chr_df[pos_col] >= start) & (chr_df[pos_col] < end)].copy()
                chunk_df = annotate_alleles_from_rsid_pos(
                    chunk_df, database, rsid_col, chrom_col, pos_col, a1_col, a2_col
                )
                logger.info(
                    f"Processing {chrom}:{start}-{end}, chunk No.{ith}, {len(chunk_df)} rows."
                )
                ith += 1
                if ith == 1:
                    chunk_df.to_csv(outfile, sep="\t", index=False, mode="w")
                else:
                    chunk_df.to_csv(outfile, sep="\t", index=False, mode="a", header=False)
                start = chr_df[chr_df[pos_col] > end][pos_col].min()
                end = start + chunksize


def annotate_pos_alleles_from_rsid(
    indf: pd.DataFrame,
    database: str,
    rsid_col: str,
    chrom_col: str,
    pos_col: str,
    a1_col: str,
    a2_col: str,
    chunksize: int = 1000000,
    overwrite: bool = False,
    remove_failed: bool = False,
) -> pd.DataFrame:
    """Annotate a dataframe with rsids."""
    chunk_df = indf.copy()
    tb = tabix.open(database)
    chunk_df[rsid_col] = chunk_df[rsid_col].astype(str)
    chunk_df = chunk_df[
        (chunk_df[rsid_col].str.startswith("rs")) & (chunk_df[rsid_col].str.len() >= 3)
    ]
    chunk_df["rsid_fake_pos"] = "1" + chunk_df[rsid_col].str[4:]
    chunk_df["rsid_fake_pos"] = pd.to_numeric(chunk_df["rsid_fake_pos"], errors="coerce")
    chunk_df = chunk_df[chunk_df["rsid_fake_pos"].notnull()]
    chunk_df["rsid_fake_chr"] = chunk_df[rsid_col].str[2:4]
    chunk_df["rsid_fake_chr"] = pd.to_numeric(chunk_df["rsid_fake_chr"], errors="coerce")
    chunk_df = chunk_df[chunk_df["rsid_fake_chr"].notnull()]

    if len(chunk_df) == 0:
        logging.warning('No rsids start with "rs" found in the input file.')
        return pd.DataFrame()
    chunk_df["rsid_fake_pos"] = chunk_df["rsid_fake_pos"].astype(int)
    chunk_df["rsid_fake_chr"] = chunk_df["rsid_fake_chr"].astype(int)
    chunk_df = chunk_df.sort_values(["rsid_fake_chr", "rsid_fake_pos"])
    logging.info(f"{chunk_df.shape[0]} rsids found in the input file.")
    out_df = []
    ith = 0
    for rsid_fake_chr, subdf in chunk_df.groupby("rsid_fake_chr"):
        chrom = rsid_fake_chr
        start = subdf["rsid_fake_pos"].min()
        end = start + chunksize
        while not pd.isna(start):
            df = subdf[(subdf["rsid_fake_pos"] >= start) & (subdf["rsid_fake_pos"] < end)].copy()
            ith += 1
            logging.info(f"Processing chunk No.{ith}, {len(df)} rows.")
            # df.to_csv("test1.csv", index=False)
            df["index"] = df.index
            end = df["rsid_fake_pos"].max()
            logging.info(f"Querying rsID: {chrom}:{start-1}-{end} from database.")
            # query_res = check_output(
            #     f"tabix {database} {chrom}:{start - 1}-{end}", shell=True
            # )
            # query_res = query_res.decode("utf-8")
            # query_res = pd.read_csv(
            #     StringIO(query_res),
            #     sep="\t",
            #     names=[
            #         "0",
            #         "rsid_fake_pos",
            #         "chr_dbsnp",
            #         "bp_dbsnp",
            #         "ref_dbsnp",
            #         "alt_dbsnp",
            #     ],
            # )
            query_res = pd.DataFrame(
                columns=[
                    "rsid_fake_chr",
                    "rsid_fake_pos",
                    "rsid_dbsnp",
                    "chr_dbsnp",
                    "bp_dbsnp",
                    "ref_dbsnp",
                    "alt_dbsnp",
                ],
                data=tb.query(str(chrom), start - 1, end),
            )
            query_res = query_res.astype(
                {
                    "rsid_fake_chr": int,
                    "rsid_fake_pos": int,
                    "chr_dbsnp": str,
                    "bp_dbsnp": int,
                    "ref_dbsnp": str,
                    "alt_dbsnp": str,
                }
            )
            df = df.merge(query_res, how="left", on=["rsid_fake_chr", "rsid_fake_pos"])
            if overwrite:
                df[chrom_col] = df["chr_dbsnp"]
                df[pos_col] = df["bp_dbsnp"]
                df[a1_col] = df["ref_dbsnp"]
                df[a2_col] = df["alt_dbsnp"]
            else:
                df[chrom_col] = df["chr_dbsnp"].where(df[chrom_col].isnull(), df[chrom_col])
                df[pos_col] = df["bp_dbsnp"].where(df[pos_col].isnull(), df[pos_col])
                n_wrong_chrom = len(
                    df[
                        (df[chrom_col].astype(str) != df["chr_dbsnp"])
                        & (df["chr_dbsnp"].notnull())
                        & (df[chrom_col].notnull())
                    ]["rsid_dbsnp"].unique()
                )
                if n_wrong_chrom > 0:
                    logging.warning(f"found {n_wrong_chrom} rsids with different chrom")
                n_wrong_pos = len(
                    df[
                        (df[pos_col] != df["bp_dbsnp"])
                        & (df["bp_dbsnp"].notnull())
                        & (df[pos_col].notnull())
                    ]["rsid_dbsnp"].unique()
                )
                if n_wrong_pos > 0:
                    logging.warning(f"found {n_wrong_pos} rsids with different pos")
                df.loc[df[a1_col].isnull() & df[a2_col].isnull(), a1_col] = df["ref_dbsnp"]
                df.loc[df[a1_col].isnull() & df[a2_col].isnull(), a2_col] = df["alt_dbsnp"]
                df.loc[
                    (df[a1_col].isnull() & df[a2_col].notnull())
                    & ((df[a2_col] == df["ref_dbsnp"]) | (df[a2_col] == df["alt_dbsnp"])),
                    a1_col,
                ] = df["ref_dbsnp"].where(df[a2_col] == df["alt_dbsnp"], df["alt_dbsnp"])
                df.loc[
                    (df[a1_col].notnull() & df[a2_col].isnull())
                    & ((df[a1_col] == df["ref_dbsnp"]) | (df[a1_col] == df["alt_dbsnp"])),
                    a2_col,
                ] = df["alt_dbsnp"].where(df[a1_col] == df["ref_dbsnp"], df["ref_dbsnp"])
            df = df.drop_duplicates(subset=["index"])
            for col in [
                "rsid_fake_pos",
                "rsid_fake_chr",
                "rsid_dbsnp",
                "chr_dbsnp",
                "bp_dbsnp",
                "ref_dbsnp",
                "alt_dbsnp",
                "index",
            ]:
                del df[col]
            out_df.append(df)
            start = subdf[subdf["rsid_fake_pos"] > end]["rsid_fake_pos"].min()
            end = start + chunksize
    out_df = pd.concat(out_df, ignore_index=True)
    out_df[pos_col] = pd.to_numeric(out_df[pos_col], errors="coerce")
    out_df = out_df[out_df[pos_col].notnull()]
    out_df[pos_col] = out_df[pos_col].astype(int)
    out_df = out_df.sort_values([chrom_col, pos_col])
    if remove_failed:
        out_df = out_df[
            (out_df[chrom_col].notnull())
            & (out_df[pos_col].notnull())
            & (out_df[a1_col].notnull())
            & (out_df[a2_col].notnull())
        ]
    return out_df


def annotate_pos_alleles_from_rsid_file(
    infile: str,
    outfile: str,
    database: str,
    chunksize: int = 2000000,
    rsid_col: str = ColName.RSID,
    chrom_col: str = ColName.CHR,
    pos_col: str = ColName.BP,
    a1_col: str = ColName.EA,
    a2_col: str = ColName.NEA,
    overwrite: bool = False,
) -> None:
    """Annotate a file with rsids."""
    logging.info("Loading input file...")
    indf = pd.read_csv(infile, sep="\t")
    logging.info(f"{indf.shape[0]} SNPs loaded.")
    out_df = annotate_pos_alleles_from_rsid(
        indf,
        database,
        rsid_col,
        chrom_col,
        pos_col,
        a1_col,
        a2_col,
        chunksize,
        overwrite,
    )
    out_df.to_csv(outfile, sep="\t", index=False)



"""Console script for smunger."""

import logging
from pathlib import Path
from enum import Enum
import typer

from smunger import __version__, console

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
app = typer.Typer(context_settings=CONTEXT_SETTINGS, add_completion=False)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    version: bool = typer.Option(False, "--version", "-V", help="Show version."),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Show verbose info."),
):
    """smunger: munger for GWAS summary statistics."""
    console.rule("[bold blue]smunger[/bold blue]")
    console.print(f"Version: {__version__}", justify="center")
    console.print("Author: Jianhua Wang", justify="center")
    console.print("Email: jianhua.mert@gmail.com", justify="center")
    if version:
        typer.echo(f"smunger version: {__version__}")
        raise typer.Exit()
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.info("Verbose mode is on.")
    else:
        logging.getLogger().setLevel(logging.INFO)


@app.command()
def mapheader(
    infile: str = typer.Argument(..., help="Input summary statistics."),
    outfile: Path = typer.Argument(..., help="Output file, json."),
    sep: str = typer.Option(None, "--sep", "-s", help="Separator of the input file."),
    nrows: int = typer.Option(5, "--nrows", "-n", help="Number of rows to display."),
    skiprows: int = typer.Option(0, "--skiprows", "-k", help="Number of rows to skip."),
    comment: str = typer.Option(None, "--comment", "-c", help="Comment character."),
    gzipped: bool = typer.Option(
        None, "--gzipped", "-z", help="Input file is gzipped."
    ),
):
    """Map column names."""
    from smunger.io import load_sumstats
    from smunger.mapheader import map_colnames

    df = load_sumstats(
        infile,
        nrows=nrows,
        sep=sep,
        skiprows=skiprows,
        comment=comment,
        gzipped=gzipped,
    )
    map_colnames(df, outfile=outfile)


@app.command()
def munge(
    infile: str = typer.Argument(..., help="Input summary statistics."),
    outfile: str = typer.Argument(..., help="Output munged summary statistics."),
    colmap: str = typer.Argument(..., help="Column map file, json."),
    sep: str = typer.Option(None, "--sep", "-s", help="Separator of the input file."),
    skiprows: int = typer.Option(0, "--skiprows", "-k", help="Number of rows to skip."),
    comment: str = typer.Option(None, "--comment", "-c", help="Comment character."),
    gzipped: bool = typer.Option(
        None, "--gzipped", "-z", help="Input file is gzipped."
    ),
    build_index: bool = typer.Option(
        True, "--build-index", "-b", help="Build tabix index."
    ),
    sigsnps: str = typer.Option(
        None, "--sigsnps", "-S", help="save significant SNPs to file."
    ),
    sigsnps_pval: float = typer.Option(
        5e-8, "--sigsnps-pval", "-P", help="p-value threshold for significant SNPs."
    ),
    report: str = typer.Option(None, "--report", "-R", help="save report to file."),
):
    """Munge summary statistics."""
    import smunger
    from smunger.io import load_sumstats, save_sumstats

    df = load_sumstats(
        infile, sep=sep, skiprows=skiprows, comment=comment, gzipped=gzipped
    )
    pre_nrow = len(df)
    df = smunger.extract_cols(df, colname_map=colmap)
    df = smunger.munge(df)
    after_nrow = len(df)
    save_sumstats(df, outfile, build_index=build_index)
    from smunger.smunger import get_sigdf

    df_sig = get_sigdf(df, pval=sigsnps_pval)
    if sigsnps:
        if len(df_sig) > 0:
            save_sumstats(df_sig, sigsnps, build_index=False)
        else:
            console.print("[bold red]No significant SNPs found.[/bold red]")
    non_null_cols = df.columns[df.notnull().any()].tolist()
    report_json = {
        "in_rows": pre_nrow,
        "out_rows": after_nrow,
        "sigsnps": len(df_sig),
        "non_null_cols": non_null_cols,
    }
    if report:
        import json

        with open(report, "w") as f:
            json.dump(report_json, f, indent=4)


@app.command()
def extract(
    infile: str = typer.Argument(..., help="Input summary statistics."),
    outfile: str = typer.Argument(..., help="Output munged summary statistics."),
    rename_headers: str = typer.Option(
        None, "--rename-headers", "-r", help="Rename headers, json."
    ),
    chrom: int = typer.Option(None, "--chrom", "-c", help="Chromosome."),
    start: int = typer.Option(None, "--start", "-s", help="Start position."),
    end: int = typer.Option(None, "--end", "-e", help="End position."),
    no_bgzip: bool = typer.Option(
        True, "--no-bgzip", "-z", help="Do not bgzip output file."
    ),
):
    """Extract columns."""
    from smunger.io import export_sumstats
    import json

    if rename_headers:
        headers_map = json.loads(rename_headers)
    else:
        headers_map = None
    export_sumstats(
        filename=infile,
        out_filename=outfile,
        rename_headers=headers_map,
        chrom=chrom,
        start=start,
        end=end,
        bgzipped=no_bgzip,
    )


class Build(str, Enum):
    """Genome Builds."""

    hg17 = "hg17"
    hg18 = "hg18"
    hg19 = "hg19"
    hg38 = "hg38"


@app.command()
def liftover(
    infile: str = typer.Argument(..., help="Input summary statistics."),
    outfile: str = typer.Argument(..., help="Output munged summary statistics."),
    inbuild: Build = typer.Option(..., "--inbuild", "-i", help="Input build."),
    outbuild: Build = typer.Option(..., "--outbuild", "-o", help="Output build."),
    chromcol: str = typer.Option("CHR", "--chromcol", "-C", help="chromosome column."),
    poscol: str = typer.Option("BP", "--poscol", "-p", help="position column."),
):
    """Liftover summary statistics."""
    from smunger.liftover import liftover_file

    liftover_file(infile, outfile, inbuild, outbuild, chromcol, poscol)


@app.command()
def annorsid(
    infile: str = typer.Argument(..., help="Input summary statistics."),
    outfile: str = typer.Argument(..., help="Output munged summary statistics."),
    database: str = typer.Option(..., "--database", "-d", help="Database."),
    chunksize: int = typer.Option(2000000, "--chunksize", "-c", help="Chunk size."),
    rsidcol: str = typer.Option("rsID", "--rsidcol", "-r", help="rsid column."),
    chromcol: str = typer.Option("CHR", "--chromcol", "-C", help="chromosome column."),
    poscol: str = typer.Option("BP", "--poscol", "-p", help="position column."),
    eacol: str = typer.Option("EA", "--eacol", "-e", help="effect allele column."),
    neacol: str = typer.Option(
        "NEA", "--neacol", "-n", help="non-effect allele column."
    ),
):
    """Annotate rsid."""
    from smunger.annotate import annotate_rsid_file

    annotate_rsid_file(
        infile, outfile, database, chunksize, rsidcol, chromcol, poscol, eacol, neacol
    )


@app.command()
def annoallele(
    infile: str = typer.Argument(..., help="Input summary statistics."),
    outfile: str = typer.Argument(..., help="Output munged summary statistics."),
    database: str = typer.Option(..., "--database", "-d", help="Database."),
    chunksize: int = typer.Option(2000000, "--chunksize", "-c", help="Chunk size."),
    rsidcol: str = typer.Option("rsID", "--rsidcol", "-r", help="rsid column."),
    chromcol: str = typer.Option("CHR", "--chromcol", "-C", help="chromosome column."),
    poscol: str = typer.Option("BP", "--poscol", "-p", help="position column."),
    a1col: str = typer.Option("EA", "--a1col", "-a", help="effect allele column."),
    a2col: str = typer.Option(
        "NEA", "--a2col", "-b", help="non-effect allele column."
    ),
):
    """Annotate coordinates and alleles according to rsid."""
    from smunger.annotate import annotate_alleles_from_rsid_pos_file

    annotate_alleles_from_rsid_pos_file(
        infile,
        outfile,
        database,
        chunksize,
        rsidcol,
        chromcol,
        poscol,
        a1col,
        a2col,
    )


if __name__ == "__main__":
    app()



"""Read and write data from/to files."""

import gzip
import logging
import shutil
import os
from pathlib import Path
from subprocess import PIPE, run
from typing import Optional

import pandas as pd
import tabix
from .constant import ColName
from .smunger import munge

logger = logging.getLogger('io')


def load_sumstats(
    filename: str,
    sep: Optional[str] = None,
    nrows: Optional[int] = None,
    skiprows: int = 0,
    comment: Optional[str] = None,
    gzipped: Optional[bool] = None,
) -> pd.DataFrame:
    """Load summary statistics from a file."""
    # determine whether the file is gzipped
    if gzipped is None:
        gzipped = filename.endswith('gz')

    # read the first line of the file to determine the separator
    if sep is None:
        if gzipped:
            with gzip.open(filename, 'rt') as f:
                line = f.readline()
        else:
            with open(filename, 'rt') as f:
                line = f.readline()
        if '\t' in line:
            sep = '\t'
        elif ',' in line:
            sep = ','
        else:
            sep = ' '
    logger.info(f'File {filename} is gzipped: {gzipped}')
    logger.info(f'Separator is {sep}')
    logger.info(f'loading data from {filename}')
    # determine the separator, automatically if not specified
    return pd.read_csv(
        filename, sep=sep, nrows=nrows, skiprows=skiprows, comment=comment, compression='gzip' if gzipped else None
    )


def check_header(filename) -> bool:
    """Check if the header of a file contains the required columns."""
    header = load_sumstats(filename, nrows=5)
    if list(header.columns) != list(ColName.OUTCOLS):
        logger.error(f'Header of {filename} does not contain the required columns.')
        raise ValueError(f'Header of {filename} does not contain the required columns.')
    else:
        return True


def check_tool(tool: str) -> str:
    """
    Check if tool is installed.

    Parameters
    ----------
    tool : str
        The tool name.

    Returns
    -------
    str
        The path of the tool if it is installed, otherwise exit the program.
    """
    tool_path = shutil.which(tool)
    if tool_path:
        return tool_path
    else:
        logger.error(f"{tool} is not installed. Please install it first and make sure it is in your PATH.")
        raise ValueError(f"{tool} is not installed. Please install it first and make sure it is in your PATH.")


def save_sumstats(sumstats: pd.DataFrame, filename: str, build_index: bool = True, bgzipped: bool = True):
    """Save summary statistics to a file."""
    # save the summary statistics to a file
    filename_path = Path(filename)
    if filename_path.exists():
        logger.warning(f'File {filename} already exists. Overwriting.')

    if filename_path.suffix == '.gz':
        filename_path = filename_path.with_suffix('')
    sumstats = sumstats.sort_values(by=[ColName.CHR, ColName.BP])
    logger.info(f'Saving summary statistics to {filename_path}')
    sumstats.to_csv(filename_path, sep='\t', index=False, header=True, float_format='%g')

    # compress the file
    if bgzipped:
        compress(str(filename_path))
    # index the file
    if build_index and bgzipped:
        index(str(filename_path) + '.gz')


def compress(filename: str):
    """Compress a file with bgzip."""
    bgzip = check_tool('bgzip')
    logger.info(f'Compressing {filename} with bgzip')
    run([bgzip, '-f', str(filename)], stdout=PIPE, stderr=PIPE, check=True)


def index(filename: str, start: int = 1, end: int = 2, skip: int = 1):
    """Index a file with tabix."""
    tabix = check_tool('tabix')
    logger.info(f'Indexing {filename} with tabix')
    run(
        [tabix, '-f', '-S', str(skip), '-s', str(start), '-b', str(end), '-e', str(end), filename],
        stdout=PIPE,
        stderr=PIPE,
        check=True,
    )


def export_sumstats(
    filename: str,
    chrom: Optional[int] = None,
    start: Optional[int] = None,
    end: Optional[int] = None,
    rename_headers: Optional[dict] = None,
    out_filename: Optional[str] = None,
    bgzipped: bool = True,
) -> pd.DataFrame:
    """Export summary statistics to a file."""
    if chrom and start and end:
        logger.info(f'Loading summary statistics from {filename} for {chrom}:{start}-{end}')
        if not os.path.exists(filename + '.tbi'):
            raise FileNotFoundError(f'Index file {filename}tbi does not exist. Please index the file first.')
        else:
            tb = tabix.open(filename)
            indf = pd.DataFrame(columns=ColName.OUTCOLS, data=tb.query(str(chrom), start, end))
            indf = munge(indf)
    else:
        logger.info(f'Loading summary statistics from {filename}')
        indf = load_sumstats(filename)
    if rename_headers:
        logger.info(f'Renaming headers to {rename_headers}')
        indf = indf[list(rename_headers.keys())].copy()
        indf = indf.rename(columns=rename_headers)
    if out_filename:
        save_sumstats(indf, out_filename, build_index=False, bgzipped=bgzipped)
    return indf
